from zodipy._model import model_registry
from zodipy import components
from zodipy import source_params


model_registry.register_model(
    name="dirbe",
    components=components.DIRBE,
    spectrum=source_params.SPECTRUM_DIRBE,
    emissivities=source_params.EMISSIVITY_DIRBE,
    albedos=source_params.ALBEDO_DIRBE,
    phase_coefficients=source_params.PHASE_FUNCTION_DIRBE,
)


model_registry.register_model(
    name="planck13",
    components=components.DIRBE,
    spectrum=source_params.SPECTRUM_PLANCK,
    emissivities=source_params.EMISSIVITY_PLANCK_13,
)


model_registry.register_model(
    name="planck15",
    components=components.PLANCK,
    spectrum=source_params.SPECTRUM_PLANCK,
    emissivities=source_params.EMISSIVITY_PLANCK_15,
)


model_registry.register_model(
    name="planck18",
    components=components.PLANCK,
    spectrum=source_params.SPECTRUM_PLANCK,
    emissivities=source_params.EMISSIVITY_PLANCK_18,
)

model_registry.register_model(
    name="odegard",
    components=components.PLANCK,
    spectrum=source_params.SPECTRUM_PLANCK,
    emissivities=source_params.EMISSIVITY_ODEGARD,
)



from __future__ import annotations
from dataclasses import dataclass, field

from astropy.units import Quantity
import astropy.units as u
import numpy as np

from zodipy._component import Component
from zodipy._component_label import ComponentLabel
from zodipy.source_params import T_0_DIRBE, DELTA_DIRBE


@dataclass
class Model:
    name: str
    components: dict[ComponentLabel, Component]
    spectrum: Quantity[u.Hz] | Quantity[u.m]
    emissivities: dict[ComponentLabel, tuple[float, ...]]
    albedos: dict[ComponentLabel, tuple[float, ...]] | None = None
    phase_coefficients: list[tuple[float, ...]] | None = None
    T_0: float = T_0_DIRBE
    delta: float = DELTA_DIRBE

    @property
    def n_components(self) -> int:
        return len(self.components)

    def get_extrapolated_component_parameters(
        self,
        component_label: ComponentLabel,
        frequency: Quantity[u.GHz] | Quantity[u.m],
    ) -> tuple[float, float, tuple[float, float, float]]:
        """Returns interpolated/extrapolated spectral parameters given a frequency."""

        frequency = frequency.to(self.spectrum.unit, equivalencies=u.spectral())

        emissivity = np.interp(
            frequency, self.spectrum, self.emissivities[component_label]
        )
        if self.albedos is not None:
            albedo = np.interp(frequency, self.spectrum, self.albedos[component_label])
        else:
            albedo = 0.0

        if self.phase_coefficients is not None:
            phase_coefficient = [
                np.interp(frequency, self.spectrum, coeff)
                for coeff in self.phase_coefficients
            ]
        else:
            phase_coefficient = [0.0 for _ in range(3)]

        return emissivity, albedo, tuple(phase_coefficient)


@dataclass
class ModelRegistry:
    """Container for registered models."""

    _registry: dict[str, Model] = field(init=False, repr=False, default_factory=dict)

    @property
    def models(self) -> list[str]:
        """Returns a list of registered model names."""

        return list(self._registry.keys())

    def register_model(
        self,
        name: str,
        components: dict[ComponentLabel, Component],
        spectrum: Quantity[u.Hz] | Quantity[u.m],
        emissivities: dict[ComponentLabel, tuple[float, ...]],
        albedos: dict[ComponentLabel, tuple[float, ...]] | None = None,
        phase_coefficients: list[tuple[float, ...]] | None = None,
        T_0: float = T_0_DIRBE,
        delta: float = DELTA_DIRBE,
    ) -> None:
        """Registers a new model.

        For an example of how to register a custom model, see the documentation
        on https://zodipy.readthedocs.io/en/latest/ (coming soon).

        Parameters
        ----------
        name
            String representing the name of the model. This is the name that is
            used for the 'model' argument when initializing `Zodipy`.
        components
            Dict mapping `CompLabel`s to `Component` classes.
        spectrum
            The spectrum (frequency or length units) corresponding to the
            frequencies used to estimate the spectral parameters.
        emissivities
            Emissivity factor fits for each IPD component at the frequencies
            corresponding to 'spectrum'.
        albedos
            Albedo factor fits for each IPD component at the frequencies
            corresponding to 'spectrum'.
        phase_coefficients
            Coefficient fits for the phase function at the frequencies
            corresponding to 'spectrum'.
        T_0
            Interplanetary temperature at 1 AU. Defaults to the DIRBE model value.
        delta
            Interplanetary temperatue powerlaw parameter describing how the
            interplanetary temperature falls with radial distance from the Sun.
            Defaults to the DIRBE model value.
        """
        
        if (name := name.lower()) in self._registry:
            raise ValueError(f"a model by the name {name!s} is already registered.")

        self._registry[name] = Model(
            name=name,
            components=components,
            spectrum=spectrum,
            emissivities=emissivities,
            T_0=T_0,
            delta=delta,
            albedos=albedos,
            phase_coefficients=phase_coefficients,
        )

    def get_model(self, name: str) -> Model:
        """Returns a registered model given a name."""

        if (name := name.lower()) not in self._registry:
            raise ModuleNotFoundError(
                f"{name} is not a model in the registry. Avaliable models are: "
                f"{', '.join(self._registry)}"
            )

        return self._registry[name]


model_registry = ModelRegistry()



from __future__ import annotations

import numpy as np
from numpy.typing import NDArray

from zodipy._model import Model
from zodipy.models import model_registry


DEFAULT_EARTH_POS = np.array([1.0, 0.0, 0.0])


def tabulate_density(
    grid: NDArray[np.floating] | list[NDArray[np.floating]],
    model: str | Model = "DIRBE",
    earth_position: NDArray[np.floating] = DEFAULT_EARTH_POS,
) -> NDArray[np.floating]:
    """Returns the tabulated densities of the Interplanetary Dust components.

    Parameters
    ----------
    grid
        A cartesian mesh grid (x, y, z) created with `np.meshgrid` for which to
        tabulate the Interplanetary dust components.
    model
        The model who's Interplanetary Dust components to tabulate.
    earth_position
        The position of the Earth.

    Returns
    -------
    density_grid
        The tabulate densities of the Interplanetary Dust components.
    """

    if not isinstance(model, Model):
        model = model_registry.get_model(model)

    if not isinstance(grid, np.ndarray):
        grid = np.asarray(grid)

    earth_position = np.reshape(earth_position, (3, 1, 1, 1))

    density_grid = np.zeros((model.n_components, *grid.shape[1:]))
    for idx, comp in enumerate(model.components.values()):
        comp.X_0 = np.reshape(comp.X_0, (3, 1, 1, 1))
        density_grid[idx] = comp.compute_density(
            X_helio=grid,
            X_earth=earth_position,
        )
        comp.X_0 = np.reshape(comp.X_0, (3, 1))

    return density_grid



from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass

import numpy as np
from numpy.typing import NDArray

π = np.pi


@dataclass
class Component(ABC):
    """Base class for a DIRBE Interplanetary Dust Component.

    Any component that inherits from this class needs to implement the
    `compute_density` method which takes in the heliocentric ecliptic cartesian
    coordinates of the line of sight, and optionally Earth's position.

    Parameters
    ----------
    x_0
        x-offset from the Sun in heliocentric ecliptic coordinates [AU].
    y_0
        y-offset from the Sun in heliocentric ecliptic coordinates [AU].
    z_0
        z-offset from the Sun in heliocentric ecliptic coordinates [AU].
    i
        Inclination with respect to the ecliptic plane [deg].
    Omega
        Ascending node [deg].
    """

    x_0: float
    y_0: float
    z_0: float
    i: float
    Omega: float

    def __post_init__(self) -> None:
        # Offset vector
        self.X_0 = np.expand_dims([self.x_0, self.y_0, self.z_0], axis=1)

        # Frequently used quantities
        self.sin_i_rad = np.sin(np.radians(self.i))
        self.cos_i_rad = np.cos(np.radians(self.i))
        self.sin_Omega_rad = np.sin(np.radians(self.Omega))
        self.cos_Omega_rad = np.cos(np.radians(self.Omega))

    @abstractmethod
    def compute_density(
        self,
        X_helio: NDArray[np.floating],
        *,
        X_earth: NDArray[np.floating] | None = None,
    ) -> NDArray[np.floating]:
        """Returns the dust density of a component at points in the Solar System
        given by 'X_helio'.

        Parameters
        ----------
        X_helio
            Heliocentric ecliptic coordinates (x, y, z) of points in the Solar
            System [AU].
        X_earth
            Heliocentric ecliptic coordinates of the Earth. Required for the
            Earth-trailing Feature [AU].

        Returns
        -------
            Density of the component at the points given by 'X_helio' [1/AU].
        """


@dataclass
class Cloud(Component):
    """Diffuse Cloud component.

    Parameters
    ----------
    n_0
        Density at 1 AU.
    alpha
        Radial power-law exponent.
    beta
       Vertical shape parameter.
    gamma
        Vertical power-law exponent.
    mu
        Widening parameter for the modified fan.
    """

    n_0: float
    alpha: float
    beta: float
    gamma: float
    mu: float

    def compute_density(
        self, X_helio: NDArray[np.floating], **_
    ) -> NDArray[np.floating]:
        """See base class for documentation."""

        X_cloud = X_helio - self.X_0
        R_cloud = np.sqrt(X_cloud[0] ** 2 + X_cloud[1] ** 2 + X_cloud[2] ** 2)

        Z_cloud = (
            X_cloud[0] * self.sin_Omega_rad * self.sin_i_rad
            - X_cloud[1] * self.cos_Omega_rad * self.sin_i_rad
            + X_cloud[2] * self.cos_i_rad
        )

        ζ = np.abs(Z_cloud / R_cloud)
        μ = self.mu
        g = np.zeros_like(ζ)

        condition = ζ < μ
        g[condition] = ζ[condition] ** 2 / (2 * μ)
        g[~condition] = ζ[~condition] - (μ / 2)

        return self.n_0 * R_cloud ** -self.alpha * np.exp(-self.beta * g ** self.gamma)


@dataclass
class Band(Component):
    """Dust Band component.

    Parameters
    ----------
    n_0
        Density at 3 AU.
    delta_zeta
        Shape parameter [deg].
    v
        Shape parameter.
    p
        Shape parameter.
    delta_r
        Inner radial cutoff.
    """

    n_0: float
    delta_zeta: float
    v: float
    p: float
    delta_r: float

    def __post_init__(self) -> None:
        super().__post_init__()
        self.delta_zeta_rad = np.radians(self.delta_zeta)

    def compute_density(
        self, X_helio: NDArray[np.floating], **_
    ) -> NDArray[np.floating]:
        """See base class for documentation."""

        X_band = X_helio - self.X_0
        R_band = np.sqrt(X_band[0] ** 2 + X_band[1] ** 2 + X_band[2] ** 2)

        Z_band = (
            X_band[0] * self.sin_Omega_rad * self.sin_i_rad
            - X_band[1] * self.cos_Omega_rad * self.sin_i_rad
            + X_band[2] * self.cos_i_rad
        )

        ζ = np.abs(Z_band / R_band)
        ζ_over_δ_ζ = ζ / self.delta_zeta_rad
        term1 = 3 * self.n_0 / R_band
        term2 = np.exp(-(ζ_over_δ_ζ ** 6))

        # Differs from eq 8 in K98 by a factor of 1/self.v. See Planck XIV
        # section 4.1.2.
        term3 = 1 + (ζ_over_δ_ζ ** self.p) / self.v

        term4 = 1 - np.exp(-((R_band / self.delta_r) ** 20))

        return term1 * term2 * term3 * term4


@dataclass
class Ring(Component):
    """Circum-solar Ring component.

    Parameters
    ----------
    n_0
        Density at 1 AU.
    R
        Radius of the peak density.
    sigma_r
        Radial dispersion.
    sigma_z
        Vertical dispersion.
    """

    n_0: float
    R: float
    sigma_r: float
    sigma_z: float

    def compute_density(
        self, X_helio: NDArray[np.floating], **_
    ) -> NDArray[np.floating]:
        """See base class for documentation."""

        X_ring = X_helio - self.X_0
        R_ring = np.sqrt(X_ring[0] ** 2 + X_ring[1] ** 2 + X_ring[2] ** 2)

        Z_ring = (
            X_ring[0] * self.sin_Omega_rad * self.sin_i_rad
            - X_ring[1] * self.cos_Omega_rad * self.sin_i_rad
            + X_ring[2] * self.cos_i_rad
        )
        # Differs from eq 9 in K98 by a factor of 1/2 in the first and last
        # term. See Planck 2013 XIV, section 4.1.3.
        term1 = -((R_ring - self.R) ** 2) / self.sigma_r ** 2
        term2 = np.abs(Z_ring) / self.sigma_z

        return self.n_0 * np.exp(term1 - term2)


@dataclass
class Feature(Component):
    """Earth-trailing Feature component.

    Parameters
    ----------
    n_0
        Density at 1 AU.
    R
        Radius of the peak density.
    sigma_r
        Radial dispersion.
    sigma_z
        Vertical dispersion.
    theta
        Longitude with respect to Earth [deg].
    sigma_theta
        Longitude dispersion [deg].
    """

    n_0: float
    R: float
    sigma_r: float
    sigma_z: float
    theta: float
    sigma_theta: float

    def __post_init__(self) -> None:
        super().__post_init__()
        self.theta_rad = np.radians(self.theta)
        self.sigma_theta_rad = np.radians(self.sigma_theta)

    def compute_density(
        self,
        X_helio: NDArray[np.floating],
        X_earth: NDArray[np.floating],
        **_,
    ) -> NDArray[np.floating]:
        """See base class for documentation."""

        X_feature = X_helio - self.X_0
        R_feature = np.sqrt(X_feature[0] ** 2 + X_feature[1] ** 2 + X_feature[2] ** 2)

        Z_feature = (
            X_feature[0] * self.sin_Omega_rad * self.sin_i_rad
            - X_feature[1] * self.cos_Omega_rad * self.sin_i_rad
            + X_feature[2] * self.cos_i_rad
        )
        X_earth_comp = X_earth - self.X_0

        θ_comp = np.arctan2(X_feature[1], X_feature[0]) - np.arctan2(
            X_earth_comp[1], X_earth_comp[0]
        )

        Δθ = θ_comp - self.theta_rad
        condition1 = Δθ < -π
        condition2 = Δθ > π
        Δθ[condition1] = Δθ[condition1] + 2 * π
        Δθ[condition2] = Δθ[condition2] - 2 * π

        # Differs from eq 9 in K98 by a factor of 1/2 in the first and last
        # term. See Planck 2013 XIV, section 4.1.3.
        exp_term = (R_feature - self.R) ** 2 / self.sigma_r ** 2
        exp_term += np.abs(Z_feature) / self.sigma_z
        exp_term += Δθ ** 2 / self.sigma_theta_rad ** 2

        return self.n_0 * np.exp(-exp_term)



from __future__ import annotations
from typing import Sequence

from astropy.units import Quantity
import astropy.units as u
import healpy as hp
import numpy as np
from numpy.typing import NDArray


def get_unit_vector_from_pixels(
    coord_in: str, pixels: Sequence[int] | NDArray[np.integer], nside: int
) -> NDArray[np.floating]:
    """Returns ecliptic unit vectors from HEALPix pixels representing some pointing."""

    unit_vectors = np.asarray(hp.pix2vec(nside, pixels))

    return np.asarray(hp.Rotator(coord=[coord_in, "E"])(unit_vectors))


def get_unit_vector_from_angles(
    coord_in: str,
    phi: Quantity[u.rad] | Quantity[u.deg],
    theta: Quantity[u.rad] | Quantity[u.deg],
    lonlat: bool = False,
) -> NDArray[np.floating]:
    """Returns ecliptic unit vectors from sky angles representing some pointing."""

    unit_vectors = np.asarray(hp.ang2vec(theta, phi, lonlat=lonlat)).transpose()

    return np.asarray(hp.Rotator(coord=[coord_in, "E"])(unit_vectors))



from __future__ import annotations

import numpy as np
from numpy.typing import NDArray

from zodipy._component import Component
from zodipy._source import (
    get_blackbody_emission_nu,
    get_interplanetary_temperature,
    get_phase_function,
    get_solar_flux,
)


def get_emission_step(
    r: float | NDArray[np.floating],
    *,
    frequency: float,
    observer_position: NDArray[np.floating],
    earth_position: NDArray[np.floating],
    unit_vectors: NDArray[np.floating],
    component: Component,
    T_0: float,
    delta: float,
    emissivity: float,
    albedo: float,
    phase_coefficients: tuple[float, float, float],
) -> NDArray[np.floating]:
    """Returns the Zodiacal emission at a step along a line of sight.

    This function computes equation (1) in Kelsall et al. (1998) along a step in ds.

    Parameters
    ----------
    r
        Radial distance along the line-of-sight [AU / 1 AU].
    frequency
        Frequency at which to evaluate the brightness integral [GHz].
    observer_position
        The heliocentric ecliptic cartesian position of the observer [AU / 1 AU].
    earth_position
        The heliocentric ecliptic cartesian position of the Earth [AU / 1 AU].
    unit_vectors
        Heliocentric ecliptic cartesian unit vectors for each pointing [AU / 1 AU].
    component
        Interplanetary Dust component.
    T_0
        Interplanetary temperature at 1 AU.
    delta
        Interplanetary temperature power law parameter.
    emissivity
        Frequency and component dependant emissivity factor representing the 
        deviation of the emisstion from a black body.
    albedo
        Frequency and component dependant albedo factor representing the
        probability of scatterin.
    phase_coefficient
        Frequency dependant parameters representing the distribution of scattered
        emission.

    Returns
    -------
        The Zodiacal emission from an Interplanetary Dust component at a step 
        along line of sights in units of W / Hz / m^2 / sr.
    """

    r_vec = r * unit_vectors
    X_helio = r_vec + observer_position
    R_helio = np.sqrt(X_helio[0] ** 2 + X_helio[1] ** 2 + X_helio[2] ** 2)

    density = component.compute_density(X_helio=X_helio, X_earth=earth_position)
    interplanetary_temperature = get_interplanetary_temperature(R_helio, T_0, delta)
    blackbody_emission = get_blackbody_emission_nu(
        frequency, interplanetary_temperature
    )

    if albedo > 0:
        emission = (1 - albedo) * (emissivity * blackbody_emission)
        scattering_angle = np.arccos(np.sum(r_vec * X_helio, axis=0) / (r * R_helio))
        solar_flux = get_solar_flux(R_helio, frequency)
        phase_function = get_phase_function(scattering_angle, phase_coefficients)
        emission += albedo * solar_flux * phase_function

    else:
        emission = emissivity * blackbody_emission

    return emission * density



from __future__ import annotations
from functools import lru_cache

import astropy.constants as const
import astropy.units as u
import numpy as np
from numpy.typing import NDArray


h = const.h.value
c = const.c.value
k_B = const.k_B.value
R_sun = const.R_sun.to(u.AU).value
T_sun = 5778


def get_blackbody_emission_nu(
    freq: float | NDArray[np.floating],
    T: float | NDArray[np.floating],
) -> NDArray[np.floating]:
    """Returns the blackbody emission.

    Parameters
    ----------
    T
        Temperature of the blackbody [K].
    freq
        Frequency [GHz] .

    Returns
    -------
        Blackbody emission [W / m^2 Hz sr].
    """

    freq *= 1e9
    term1 = (2 * h * freq ** 3) / c ** 2
    term2 = np.expm1(((h * freq) / (k_B * T)))

    return term1 / term2


def get_solar_flux(
    R: NDArray[np.floating], freq: float, T: float = T_sun
) -> NDArray[np.floating]:
    """Returns the solar flux observed at some distance R from the Sun in AU.

    Parameteers
    -----------
    freq
        Frequency [GHz].
    T
        Temperature [K].

    Returns
    -------
        Solar flux at some distance R from the Sun in AU.
    """

    return np.pi * get_blackbody_emission_nu(T=T, freq=freq) * (R_sun / R) ** 2


def get_interplanetary_temperature(
    R: NDArray[np.floating], T_0: float, delta: float
) -> NDArray[np.floating]:
    """Returns the Interplanetary Temperature given a radial distance from the Sun.

    Parameters
    ----------
    R
        Radial distance from the sun in ecliptic heliocentric coordinates [AU / 1AU].
    T_0
        Temperature of the solar system at 1 AU [K].
    delta
        Powerlaw index.

    Returns
    -------
        Interplanetary temperature [K].
    """

    return T_0 * R ** -delta


def get_phase_function(
    Theta: NDArray[np.floating], phase_coefficients: tuple[float, float, float]
) -> NDArray[np.floating]:
    """Returns the phase function.

    Parameters
    ----------
    Theta
        Scattering angle.
    coeffs
        Phase function parameters.

    Returns
    -------
        The Phase funciton.
    """

    c_0, c_1, c_2 = phase_coefficients
    phase_normalization = _get_phase_normalization(c_0, c_1, c_2)

    return phase_normalization * (c_0 + c_1 * Theta + np.exp(c_2 * Theta))


@lru_cache
def _get_phase_normalization(c_0: float, c_1: float, c_2: float) -> float:
    """Returns the analyitcal integral for the phase normalization factor N."""

    pi = np.pi
    int_term1 = 2 * pi
    int_term2 = 2 * c_0
    int_term3 = pi * c_1
    int_term4 = (np.exp(c_2 * pi) + 1) / (c_2 ** 2 + 1)

    return 1 / (int_term1 * (int_term2 + int_term3 + int_term4))



from enum import Enum

class ComponentLabel(Enum):
    """Labels representing the Interplanetary Dust components in the DIRBE model."""

    CLOUD = "cloud"
    BAND1 = "band1"
    BAND2 = "band2"
    BAND3 = "band3"
    RING = "ring"
    FEATURE = "feature"



from zodipy.zodipy import Zodipy
from zodipy.models import model_registry
from zodipy._contour import tabulate_density
from zodipy._component_label import ComponentLabel
from zodipy import components
from zodipy import source_params


__all__ = (
    "Zodipy",
    "model_registry",
    "tabulate_density",
    "ComponentLabel",
    "components",
    "source_params",
)



from __future__ import annotations
from typing import Callable

import numpy as np
from numpy.typing import NDArray


def trapezoidal_regular_grid(
    get_emission_step: Callable[[float | NDArray[np.floating]], NDArray[np.floating]],
    start: float,
    stop: float | NDArray[np.floating],
    n_steps: int,
) -> NDArray[np.floating]:
    """
    Integrates and returns the Zodiacal Emission of a Interplanetary Dust component
    over a regular grid.

    Parameters
    ----------
    get_emission_step
        Function that computes the Zodiacal emission at a step along the line
        of sight for an Interplanetary Dust component.
    start
        Lower integration limit (At the face of the observer).
    stop
        Upper integration limit (At a distance along the line of sight which
        corresponds to a heliocentric distance of 5.2 AU)
    step
        Number of steps along the line of sight to integrate.

    Returns
    -------
    integrated_emission
        Integrated Zodiacal emission for an Interplanetary Dust component over
        line of sights in units of W / Hz / m^2 / sr.
    """
    
    ds = (stop - start) / n_steps

    integrated_emission = get_emission_step(start) + get_emission_step(stop)
    integrated_emission += 2 * sum(
        get_emission_step(start + ds * step) for step in range(1, n_steps)
    )

    return integrated_emission * (ds / 2)



from functools import partial
from typing import List, Literal, Optional, Sequence, Union

from astropy.coordinates import solar_system_ephemeris
import astropy.units as u
from astropy.units import Quantity, quantity_input
from astropy.time import Time
import healpy as hp
import numpy as np
from numpy.typing import NDArray

from zodipy._emission import get_emission_step
from zodipy._ephemeris import get_earth_position, get_observer_position
from zodipy._integral import trapezoidal_regular_grid
from zodipy._line_of_sight import get_line_of_sight
from zodipy._unit_vector import (
    get_unit_vector_from_angles,
    get_unit_vector_from_pixels,
)
from zodipy.models import model_registry


__all__ = ("Zodipy",)


class Zodipy:
    """The Zodipy interface.

    Zodipy simulates the Zodiacal emission that a Solar System observer is
    predicted to see given the Kelsall et al. (1998) interplanetary dust model.
    """

    def __init__(self, model: str = "DIRBE", ephemeris: str = "de432s") -> None:
        """Initializes Zodipy for a given a model and emphemeris.

        Parameters
        ----------
        model
            The name of the interplanetary dust model. Defaults to DIRBE. See all
            available models with `zodipy.MODELS`.
        ephemeris
            Ephemeris used to compute the positions of the observer and Earth.
            Defaults to 'de432s' which requires downloading (and caching) a 10
            MB file. For more information on available ephemeridis, please visit
            https://docs.astropy.org/en/stable/coordinates/solarsystem.html
        """

        self.model = model_registry.get_model(model)
        solar_system_ephemeris.set(ephemeris)

    @property
    def supported_observers(self) -> List[str]:
        """Returns all observers suported by the ephemeridis."""

        return list(solar_system_ephemeris.bodies) + ["l2 or semb-l2"]

    @quantity_input
    def get_emission(
        self,
        freq: Union[Quantity[u.Hz], Quantity[u.m]],
        *,
        obs: str = "earth",
        obs_time: Time = Time.now(),
        obs_pos: Optional[Quantity[u.AU]] = None,
        pixels: Optional[Union[int, Sequence[int], NDArray[np.integer]]] = None,
        theta: Optional[Union[Quantity[u.rad], Quantity[u.deg]]] = None,
        phi: Optional[Union[Quantity[u.rad], Quantity[u.deg]]] = None,
        nside: Optional[int] = None,
        lonlat: bool = False,
        binned: bool = False,
        return_comps: bool = False,
        coord_in: Literal["E", "G", "C"] = "E",
    ) -> Quantity[u.MJy / u.sr]:
        """Returns simulated Zodiacal Emission.

        This function takes as arguments a frequency or wavelength (`freq`),
        time of observation (`obs_time`), an a Solar System observer (`obs`).
        The position of the observer is computed using the pehemeris specified
        in the initialization of `Zodipy`. Optionally, the observer position
        can be explicitly specified with the `obs_pos` argument, which
        overrides`obs`. The pointing, for which to compute the emission, is
        specified either in the form of angles on the sky (`theta`, `phi`), or
        as HEALPIX pixels at some resolution (`pixels`, `nside`).

        Parameters
        ----------
        freq
            Frequency or wavelength at which to evaluate the Zodiacal emission.
            Must have units compatible with Hz or m.
        obs
            The solar system observer. A list of all support observers (for a
            given ephemeridis) is specified in `observers` attribute of the
            `zodipy.Zodipy` instance. Defaults to 'earth'.
        obs_time
            Time of observation (`astropy.time.Time` object). Defaults to
            current time.
        obs_pos
            The heliocentric ecliptic cartesian position of the observer in AU.
            Overrides the `obs` argument. Default is None.
        pixels
            A single, or a sequence of HEALPIX pixels representing points on
            the sky. The `nside` parameter, which specifies the resolution of
            these pixels, must also be provided along with this argument.
        theta, phi
            Angular coordinates (co-latitude, longitude ) of a point, or a
            sequence of points, on the sphere. Units must be radians or degrees.
            co-latitude must be in [0, pi] rad, and longitude in [0, 2*pi] rad.
        nside
            HEALPIX map resolution parameter of the pixels (and optionally the
            binned output map). Must be specified if `pixels` is provided or if
            `binned` is set to True.
        lonlat
            If True, input angles (`theta`, `phi`) are assumed to be longitude
            and latitude, otherwise, they are co-latitude and longitude.
            Seeting lonlat to True corresponds to theta=RA, phi=DEC
        return_comps
            If True, the emission is returned component-wise. Defaults to False.
        binned
            If True, the emission is binned into a HEALPIX map with resolution
            given by the `nside` argument in the coordinate frame corresponding to
            `coord_in`. Defaults to False.
        coord_in
            Coordinates frame of the pointing. Assumes 'E' (ecliptic coordinates)
            by default.

        Returns
        -------
        emission
            Sequence of simulated Zodiacal emission in units of 'MJy/sr' for
            each pointing. If the pointing is provided in a time-ordered manner,
            then the output of this function can be interpreted as the observered
            Zodiacal Emission timestream.
        """

        if obs.lower() not in self.supported_observers:
            raise ValueError(
                f"observer {obs!r} not supported by ephemeridis "
                f"{solar_system_ephemeris._value!r} or 'Zodipy'"
            )

        if pixels is not None:
            if phi is not None or theta is not None:
                raise ValueError(
                    "get_time_ordered_emission() got an argument for both 'pixels'"
                    "and 'theta' or 'phi' but can only use one of the two"
                )
            if nside is None:
                raise ValueError(
                    "get_time_ordered_emission() got an argument for 'pixels', "
                    "but argument 'nside' is missing"
                )
            if lonlat:
                raise ValueError(
                    "get_time_ordered_emission() has 'lonlat' set to True "
                    "but 'theta' and 'phi' is not given"
                )
            if np.size(pixels) == 1:
                pixels = np.expand_dims(pixels, axis=0)

        if binned and nside is None:
            raise ValueError(
                "get_time_ordered_emission() has 'binned' set to True, but "
                "argument 'nside' is missing"
            )

        if (theta is not None and phi is None) or (phi is not None and theta is None):
            raise ValueError(
                "get_time_ordered_emission() is missing an argument for either "
                "'phi' or 'theta'"
            )

        if (theta is not None) and (phi is not None):
            if theta.size != phi.size:
                raise ValueError(
                    "get_time_ordered_emission() got arguments 'theta' and 'phi' "
                    "with different size "
                )
            if theta.size == 1:
                phi = np.expand_dims(phi, axis=0)
                theta = np.expand_dims(theta, axis=0)
            if lonlat:
                theta = theta.to(u.deg)
                phi = phi.to(u.deg)
            else:
                theta = theta.to(u.rad)
                phi = phi.to(u.rad)

        if not isinstance(obs_time, Time):
            raise TypeError("argument 'obs_time' must be of type 'astropy.time.Time'")

        earth_position = get_earth_position(obs_time)
        if obs_pos is None:
            observer_position = get_observer_position(obs, obs_time, earth_position)
        else:
            observer_position = obs_pos.reshape(3,1)

        frequency = freq.to("GHz", equivalencies=u.spectral())

        if binned:
            if pixels is not None:
                unique_pixels, counts = np.unique(pixels, return_counts=True)
                unit_vectors = get_unit_vector_from_pixels(
                    coord_in=coord_in,
                    pixels=unique_pixels,
                    nside=nside,
                )
            else:
                unique_angles, counts = np.unique(
                    np.asarray([theta, phi]), return_counts=True, axis=1
                )
                unique_pixels = hp.ang2pix(nside, *unique_angles, lonlat=lonlat)
                unit_vectors = get_unit_vector_from_angles(
                    coord_in=coord_in,
                    theta=unique_angles[0],
                    phi=unique_angles[1],
                    lonlat=lonlat,
                )

            emission = np.zeros((self.model.n_components, hp.nside2npix(nside)))
            for idx, (label, component) in enumerate(self.model.components.items()):
                extrapolated_parameters = (
                    self.model.get_extrapolated_component_parameters(label, frequency)
                )
                emissivity, albedo, phase_coefficients = extrapolated_parameters

                # Create new callable with single argument as the distance along
                # the line of sight (R)
                emission_step_function = partial(
                    get_emission_step,
                    frequency=frequency.value,
                    observer_position=observer_position.value,
                    earth_position=earth_position.value,
                    unit_vectors=unit_vectors,
                    component=component,
                    T_0=self.model.T_0,
                    delta=self.model.delta,
                    emissivity=emissivity,
                    albedo=albedo,
                    phase_coefficients=phase_coefficients,
                )
                start, stop, n_steps = get_line_of_sight(
                    component_label=label,
                    observer_position=observer_position.value,
                    unit_vectors=unit_vectors,
                )
                emission[idx, unique_pixels] = trapezoidal_regular_grid(
                    get_emission_step=emission_step_function,
                    start=start,
                    stop=stop,
                    n_steps=n_steps,
                )

            emission[:, unique_pixels] *= counts

        else:
            if pixels is not None:
                unique_pixels, indicies = np.unique(pixels, return_inverse=True)
                unit_vectors = get_unit_vector_from_pixels(
                    coord_in=coord_in,
                    pixels=unique_pixels,
                    nside=nside,
                )
                emission = np.zeros((self.model.n_components, len(pixels)))

            else:
                unique_angles, indicies = np.unique(
                    np.asarray([theta, phi]), return_inverse=True, axis=1
                )
                unit_vectors = get_unit_vector_from_angles(
                    coord_in=coord_in,
                    theta=unique_angles[0],
                    phi=unique_angles[1],
                    lonlat=lonlat,
                )

                emission = np.zeros((self.model.n_components, len(theta)))

            for idx, (label, component) in enumerate(self.model.components.items()):
                extrapolated_parameters = (
                    self.model.get_extrapolated_component_parameters(label, frequency)
                )
                emissivity, albedo, phase_coefficients = extrapolated_parameters

                # Create new callable with single argument as the distance along
                # the line of sight (R)
                emission_step_function = partial(
                    get_emission_step,
                    frequency=frequency.value,
                    observer_position=observer_position.value,
                    earth_position=earth_position.value,
                    unit_vectors=unit_vectors,
                    component=component,
                    T_0=self.model.T_0,
                    delta=self.model.delta,
                    emissivity=emissivity,
                    albedo=albedo,
                    phase_coefficients=phase_coefficients,
                )
                start, stop, n_steps = get_line_of_sight(
                    component_label=label,
                    observer_position=observer_position.value,
                    unit_vectors=unit_vectors,
                )
                integrated_comp_emission = trapezoidal_regular_grid(
                    get_emission_step=emission_step_function,
                    start=start,
                    stop=stop,
                    n_steps=n_steps,
                )

                emission[idx] = integrated_comp_emission[indicies]

        # The output unit is W/Hz/m^2/sr which we convert to MJy/sr
        emission = (emission << (u.W / u.Hz / u.m ** 2 / u.sr)).to(u.MJy / u.sr)

        return emission if return_comps else emission.sum(axis=0)



from __future__ import annotations

from astropy.units import Quantity
import astropy.units as u
from astropy.time import Time
from astropy.coordinates import HeliocentricMeanEcliptic, get_body
import numpy as np


# L2 not included in the Astropy api for now so we manually implement an 
# support for it by assume that SEMB-L2 is at all times at a fixed distance 
# from Earth along the vector pointing to Earth from the Sun.
DISTANCE_FROM_EARTH_TO_L2 = Quantity(0.009896235034000056, u.AU)


def get_earth_position(time_of_observation: Time) -> Quantity[u.AU]:
    """Returns the position of the Earth given an ephemeris and observation time."""

    earth_skycoordinate = get_body("Earth", time=time_of_observation)
    earth_skycoordinate = earth_skycoordinate.transform_to(HeliocentricMeanEcliptic)
    earth_position = earth_skycoordinate.cartesian.xyz.to(u.AU)

    return earth_position.reshape(3, 1)


def get_observer_position(
    observer: str, time_of_observation: Time, earth_position: Quantity[u.AU]
) -> Quantity[u.AU]:
    """Returns the position of the Earth and the observer."""

    if observer.lower() in ("semb-l2", "l2"):
        return _get_sun_earth_moon_barycenter(earth_position)

    observer_skycoordinate = get_body("Earth", time=time_of_observation)
    observer_skycoordinate = observer_skycoordinate.transform_to(
        HeliocentricMeanEcliptic
    )
    observer_position = observer_skycoordinate.cartesian.xyz.to(u.AU)

    return observer_position.reshape(3, 1)


def _get_sun_earth_moon_barycenter(
    earth_position: Quantity[u.AU],
) -> Quantity[u.AU]:
    """Returns the approximated position of SEMB-L2 given Earth's position."""

    r_earth = np.linalg.norm(earth_position)
    earth_unit_vector = earth_position / r_earth
    semb_l2_length = r_earth + DISTANCE_FROM_EARTH_TO_L2

    return earth_unit_vector * semb_l2_length



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zodipy']

package_data = \
{'': ['*']}

install_requires = \
['astropy>=5.0.1',
 'healpy>=1.15.0,<2.0.0',
 'jplephem>=2.17,<3.0',
 'numpy>=1.22.3,<2.0.0',
 'typing-extensions>=4.1.1,<5.0.0']

setup_kwargs = {
    'name': 'zodipy',
    'version': '0.5.5',
    'description': 'Zodipy is a python tool that simulates the Zodiacal emission.',
    'long_description': '\n<img src="imgs/zodipy_logo.png" width="350">\n\n[![PyPI version](https://badge.fury.io/py/zodipy.svg)](https://badge.fury.io/py/zodipy)\n![Tests](https://github.com/MetinSa/zodipy/actions/workflows/tests.yml/badge.svg)\n[![astropy](http://img.shields.io/badge/powered%20by-AstroPy-orange.svg?style=flat)](http://www.astropy.org/)\n\n\n---\n\n\n*Zodipy* is a Python tool for simulating the Interplanetary Dust Emission that a Solar System observer sees, either in the form of timestreams or binned HEALPIX maps.\n\n![plot](imgs/zodipy_map.png)\n\n\n# Usage\nA full introduction to *Zodipy* and its use-cases can be found in the [documentation](https://zodipy.readthedocs.io/en/latest/).\n\n**Interplanetary Dust models:** select between built in models.\n```python\nfrom zodipy import Zodipy\n\nmodel = Zodipy(model="DIRBE") # DIRBE\nmodel = Zodipy(model="Planck18") # Planck 2018\n```\n\n**Get emission from a point on the sky:** choose a frequency/wavelength, an observer, a time of observation, and angular coordinates (co-latitude, longitude).\n```python\nimport astropy.units as u\nfrom astropy.time import Time\n\nmodel.get_emission(\n    25*u.micron,\n    obs="earth",\n    obs_time=Time.now(),\n    theta=10*u.deg,\n    phi=40*u.deg,\n)\n>> <Quantity [16.65684599] MJy / sr>\n```\n\n**Get emission from a sequence of angular coordinates:** `theta` and `phi` can be a sequence of angles that can represent some time-ordered pointing.\n```python\ntheta = [10.1, 10.5, 11.1, 11.5] * u.deg # Longitude\nphi = [40.2, 39.9, 39.8, 41.3] * u.deg # Latitude\n\nmodel.get_emission(\n    25*u.micron,\n    obs="earth",\n    obs_time=Time.now(),\n    theta=theta,\n    phi=phi,\n    lonlat=True,\n)\n>> <Quantity [29.11106315, 29.33735654, 29.41248579, 28.30858417] MJy / sr>\n```\n\n\n**Get emission from pixel indices on a HEALPIX grid:** a sequence of pixel indicies along with an NSIDE parameter can be used.\n```python\nmodel.get_emission(\n    25*u.micron,\n    obs="earth",\n    obs_time=Time.now(),\n    pixels=[24654, 12937, 26135],\n    nside=128,\n)\n>> <Quantity [17.77385144, 19.7889428 , 22.44797121] MJy / sr>\n```\n\n**Get binned emission component-wise:** the emission can be binned to a HEALPIX map, and also returned component-wise.\n```python\nimport healpy as hp\nimport numpy as np\n\nnside = 128\n\nmodel.get_emission(\n    25*u.micron,\n    obs="earth",\n    obs_time=Time.now(),\n    pixels=np.arange(hp.nside2npix(nside)),\n    nside=nside,\n    binned=True,\n    return_comps=True\n).shape\n>> (6, 196608)\n```\n\n# Documentation\nA detailed introduction along with a tutorial of how to use *Zodipy* will shortly be available in the [documentation](https://zodipy.readthedocs.io/en/latest/).\n# Installing\nZodipy is available on PyPI and can be installed with ``pip install zodipy`` (Python >= 3.8 required).\n\n# Scientific paper\n- San et al. (2022). *Zodipy: software for simulating Interplanetary Dust Emission.* Manuscript in preparation.\n\n\n<!-- *Zodipy* defaults to using the interplanetary dust model developed by the DIRBE team, and the `de432s` JPL ephemeris (10 MB file downloaded and cached first time `Zodipy` is initialized). The ephemeris is used to compute the position of the relevant Solar System bodies through the `astropy.coordinates.solar_system_ephemeris` api.  -->',
    'author': 'Metin San',
    'author_email': 'metinisan@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/MetinSa/zodipy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<=3.10',
}


setup(**setup_kwargs)



from __future__ import annotations

import numpy as np
from numpy.typing import NDArray

from zodipy._component_label import ComponentLabel

DISTANCE_TO_JUPITER = 5.2 # AU
EPS = np.finfo(float).eps

# Line of sight steps
line_of_sight_steps: dict[ComponentLabel, int] = {
    ComponentLabel.CLOUD: 200,
    ComponentLabel.BAND1: 200,
    ComponentLabel.BAND2: 250,
    ComponentLabel.BAND3: 200,
    ComponentLabel.RING: 125,
    ComponentLabel.FEATURE: 125,
}

# Line of sight steps
line_of_sight_cutoffs: dict[ComponentLabel, float] = {
    ComponentLabel.CLOUD: DISTANCE_TO_JUPITER,
    ComponentLabel.BAND1: DISTANCE_TO_JUPITER,
    ComponentLabel.BAND2: DISTANCE_TO_JUPITER,
    ComponentLabel.BAND3: DISTANCE_TO_JUPITER,
    ComponentLabel.RING: 3,
    ComponentLabel.FEATURE: 2,
}

def get_line_of_sight(
    component_label: ComponentLabel,
    observer_position: NDArray[np.floating],
    unit_vectors: NDArray[np.floating],
) -> tuple[float, NDArray[np.floating], int]:
    """
    Returns the start, stop and number of steps along the line of sights for an
    Interplanetary Dust component given the observer position and unit vectors
    corresponding to the pointing.
    """

    cutoff = line_of_sight_cutoffs[component_label]
    n_steps = line_of_sight_steps[component_label]
    stop = _get_line_of_sight_endpoints(cutoff, observer_position, unit_vectors)

    return EPS, stop, n_steps


def _get_line_of_sight_endpoints(
    r_cutoff: float,
    observer_position: NDArray[np.floating],
    unit_vectors: NDArray[np.floating],
) -> NDArray[np.floating]:
    """
    Computes the distance from the observer to the point along the line of
    sight which intersects the the sphere of length r_max around the Sun.
    """

    x, y, z = observer_position.flatten()

    r = np.sqrt(x ** 2 + y ** 2 + z ** 2)
    theta = np.arctan2(y, x)
    cos_theta = np.cos(theta)
    sin_theta = np.sin(theta)

    x_0 = r * cos_theta
    y_0 = r * sin_theta

    u_x, u_y, u_z = unit_vectors
    lon = np.arctan2(u_y, u_x)
    lat = np.arcsin(u_z)

    cos_lon = np.cos(lon)
    sin_lon = np.sin(lon)
    cos_lat = np.cos(lat)

    b = 2 * (x_0 * cos_lat * cos_lon + y_0 * cos_lat * sin_lon)
    c = r ** 2 - r_cutoff ** 2
    q = -0.5 * b * (1 + np.sqrt(b ** 2 - 4 * c) / np.abs(b))

    return np.maximum(q, c / q)
